// Geometric Tools, LLC
// Copyright (c) 1998-2010
// Distributed under the Boost Software License, Version 1.0.
// http://www.boost.org/LICENSE_1_0.txt
// http://www.geometrictools.com/License/Boost/LICENSE_1_0.txt
//
// File Version: 5.0.1 (2010/10/01)

#include "Wm5MathematicsPCH.h"
#include "Wm5DistPoint3Ellipsoid3.h"

namespace Wm5
{
//----------------------------------------------------------------------------
template <typename Real>
DistPoint3Ellipsoid3<Real>::DistPoint3Ellipsoid3 (const Vector3<Real>& point,
    const Ellipsoid3<Real>& ellipsoid)
    :
    mPoint(&point),
    mEllipsoid(&ellipsoid)
{
}
//----------------------------------------------------------------------------
template <typename Real>
const Vector3<Real>& DistPoint3Ellipsoid3<Real>::GetPoint () const
{
    return *mPoint;
}
//----------------------------------------------------------------------------
template <typename Real>
const Ellipsoid3<Real>& DistPoint3Ellipsoid3<Real>::GetEllipsoid ()
    const
{
    return *mEllipsoid;
}
//----------------------------------------------------------------------------
template <typename Real>
Real DistPoint3Ellipsoid3<Real>::Get ()
{
    return Math<Real>::Sqrt(GetSquared());
}
//----------------------------------------------------------------------------
template <typename Real>
Real DistPoint3Ellipsoid3<Real>::GetSquared ()
{
    // Compute coordinates of point in ellipsoid coordinate system.
    Vector3<Real> diff = *mPoint - mEllipsoid->Center;
    Vector3<Real> epoint(diff.Dot(mEllipsoid->Axis[0]),
        diff.Dot(mEllipsoid->Axis[1]), diff.Dot(mEllipsoid->Axis[2]));

    const Real* extent = mEllipsoid->Extent;
    Real a2 = extent[0]*extent[0];
    Real b2 = extent[1]*extent[1];
    Real c2 = extent[2]*extent[2];
    Real u2 = epoint.X()*epoint.X();
    Real v2 = epoint.Y()*epoint.Y();
    Real w2 = epoint.Z()*epoint.Z();
    Real a2u2 = a2*u2, b2v2 = b2*v2, c2w2 = c2*w2;

    // Initial guess for Newton's method.
    Real uRatio = epoint.X()/extent[0];
    Real vRatio = epoint.Y()/extent[1];
    Real wRatio = epoint.Z()/extent[2];
    Real t;
    if (uRatio*uRatio + vRatio*vRatio + wRatio*wRatio < (Real)1)
    {
        t = (Real)0;
    }
    else
    {
        Real maxExtent = extent[0];
        if (extent[1] > maxExtent)
        {
            maxExtent = extent[1];
        }
        if (extent[2] > maxExtent)
        {
            maxExtent = extent[2];
        }

        t = maxExtent*epoint.Length();
    }

    // Apply Newton's method.
    const int maxIteration = 64;
    Real p = (Real)1;
    Real q = (Real)1;
    Real r = (Real)1;
    for (int i = 0; i < maxIteration; ++i)
    {
        p = t + a2;
        q = t + b2;
        r = t + c2;
        Real p2 = p*p;
        Real q2 = q*q;
        Real r2 = r*r;
        Real s = p2*q2*r2 - a2u2*q2*r2 - b2v2*p2*r2 - c2w2*p2*q2;
        if (Math<Real>::FAbs(s) < Math<Real>::ZERO_TOLERANCE)
        {
            break;
        }

        Real pq = p*q, pr = p*r, qr = q*r, pqr = p*q*r;
        Real ds = ((Real)2)*(pqr*(qr + pr + pq) - a2u2*qr*(q + r)
            - b2v2*pr*(p + r) - c2w2*pq*(p + q));
        t -= s/ds;
    }

    Vector3<Real> closest(a2*epoint.X()/p, b2*epoint.Y()/q, c2*epoint.Z()/r);
    diff = closest - epoint;
    Real sqrDistance = diff.SquaredLength();

    mClosestPoint0 = *mPoint;
    mClosestPoint1 = mEllipsoid->Center +
        closest.X()*mEllipsoid->Axis[0] +
        closest.Y()*mEllipsoid->Axis[1] +
        closest.Z()*mEllipsoid->Axis[2];

    return sqrDistance;
}
//----------------------------------------------------------------------------
template <typename Real>
Real DistPoint3Ellipsoid3<Real>::Get (Real t,
    const Vector3<Real>& velocity0, const Vector3<Real>& velocity1)
{
    Vector3<Real> movedPoint = *mPoint + t*velocity0;
    Vector3<Real> movedCenter = mEllipsoid->Center + t*velocity1;
    Ellipsoid3<Real> movedEllipsoid(movedCenter, mEllipsoid->Axis,
        mEllipsoid->Extent);
    return DistPoint3Ellipsoid3<Real>(movedPoint,movedEllipsoid).Get();
}
//----------------------------------------------------------------------------
template <typename Real>
Real DistPoint3Ellipsoid3<Real>::GetSquared (Real t,
    const Vector3<Real>& velocity0, const Vector3<Real>& velocity1)
{
    Vector3<Real> movedPoint = *mPoint + t*velocity0;
    Vector3<Real> movedCenter = mEllipsoid->Center + t*velocity1;
    Ellipsoid3<Real> movedEllipsoid(movedCenter, mEllipsoid->Axis,
        mEllipsoid->Extent);
    return DistPoint3Ellipsoid3<Real>(movedPoint,movedEllipsoid).GetSquared();
}
//----------------------------------------------------------------------------

//----------------------------------------------------------------------------
// Explicit instantiation.
//----------------------------------------------------------------------------
template WM5_MATHEMATICS_ITEM
class DistPoint3Ellipsoid3<float>;

template WM5_MATHEMATICS_ITEM
class DistPoint3Ellipsoid3<double>;
//----------------------------------------------------------------------------
}
